/*
 * This file is part of SuperLib.Helper, which is an AI Library for OpenTTD
 * Copyright (C) 2011  Leif Linse
 *
 * SuperLib.Helper is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License
 *
 * SuperLib.Helper is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with SuperLib.Helper; If not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin Street, 
 * Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

import("pathfinder.road", "_SuperLib_RoadBuilder_private_RPF", 3);

class _SuperLib_RoadPathFinder_private_CustomRPF { }
function _SuperLib_RoadPathFinder_private_CustomRPF::InitializePath(sources, goals)
{
	::_SuperLib_RoadBuilder_private_RPF.InitializePath(sources, goals);
}

function _SuperLib_RoadPathFinder_private_CustomRPF::_Cost(path, new_tile, new_direction, self)
{
	local cost = ::_SuperLib_RoadBuilder_private_RPF._Cost(path, new_tile, new_direction, self);

	// Penalty for crossing railways without bridge
	if(AITile.HasTransportType(new_tile, AITile.TRANSPORT_RAIL)) cost += 800;

	// Try to build around drive through stops
	if(AIRoad.IsDriveThroughRoadStationTile(new_tile)) cost += 1000;

	if(path != null)
	{
		// Path is null on first node
		local prev_tile = path.GetTile();

		// If it is possible to reach the current tile from next tile but not in the other direction, then it is a one-way road
		// in the wrong direction.
		if(AIRoad.AreRoadTilesConnected(new_tile, prev_tile) && !AIRoad.AreRoadTilesConnected(prev_tile, new_tile))
		{
			// Don't try to use one-way roads from the back
			_SuperLib_Log.Info("One-way road detected", _SuperLib_Log.LVL_DEBUG);
			cost += 10000;
		}
	}

	return cost;
}

function _SuperLib_RoadPathFinder_private_CustomRPF::_Neighbours(path, cur_node, self)
{
	local tiles = ::_SuperLib_RoadBuilder_private_RPF._Neighbours(path, cur_node, self);

	local offsets = [AIMap.GetTileIndex(0, 1), AIMap.GetTileIndex(0, -1),
					 AIMap.GetTileIndex(1, 0), AIMap.GetTileIndex(-1, 0)];
	/* Check all tiles adjacent to the current tile. */
	foreach(offset in offsets) 
	{
		//local next_tile = cur_node + offset;
		local cur_height = AITile.GetMaxHeight(cur_node);
		if(AITile.GetSlope(cur_node) != AITile.SLOPE_FLAT) continue;

		local bridge_length = 2;
		local i_tile = cur_node + offset;
		
		while(AITile.HasTransportType(i_tile, AITile.TRANSPORT_RAIL) || AITile.IsWaterTile(i_tile)) // try to bridge over rail or flat water (rivers/canals)
		{
			i_tile += offset;
			bridge_length++;
		}

		if(bridge_length <= 2) continue; // Nothing to bridge over
		if(!Tile.IsStraight(cur_node, i_tile)) continue; // Detect map warp-arounds

		local bridge_list = AIBridgeList_Length(bridge_length);
		if(bridge_list.IsEmpty() || !AIBridge.BuildBridge(AIVehicle.VT_ROAD, bridge_list.Begin(), cur_node, i_tile)) {
			continue; // not possible to build bridge here
		}
		
		// Found possible bridge over rail
		tiles.push([i_tile, 0xFF]);
	}

	return tiles;
}

class _SuperLib_RoadPathFinder
{
	_pf = null;

	max_iterations = null;
	accumulated_iterations = null;
	step_size = null;
	find_path_error_code = null;

	// Error codes for GetFindPathError
	static PATH_FIND_NO_ERROR = 0; // No error has occured
	static PATH_FIND_FAILED_NO_PATH = 1;
	static PATH_FIND_FAILED_TIME_OUT = 2;

	constructor()
	{
		this._pf = _SuperLib_RoadBuilder_private_RPF();
	}

	/*
	 * - repair_existing = if true, use a much higher cost to build new road than for usual path finding.
	 */
	function InitializePath(sources, goals, repair_existing = false);

	/*
	 * How many iterations should at maximum be executed by PathFind in total over several calls to
	 * PathFind before it will report that no path can be found?
	 */
	function SetMaxIterations(max_iterations);

	/*
	 * How many iterations should be executed by each call to PathFind if no argument is given
	 */
	function SetStepSize(step_size);

	/*
	 * If num_iterations == null is used, SetStepSize must have been called previously to set
	 * the number of iterations to use.
	 *
	 * Returns null if either the base class FindPath returns null (no path) or it runs out of iterations.
	 * Use GetFindPathError to check what the actual error was.
	 */
	function FindPath(num_iterations = null);
	function GetFindPathError();

	/*
	 * Returns the current amount of iterations that has been used
	 */
	function GetIterationsUsed();
}

function _SuperLib_RoadPathFinder::InitializePath(sources, goals, repair_existing = false)
{
	AILog.Info("S: " + sources);
	AILog.Info("G: " + goals);
	this._pf.InitializePath(sources, goals);

	// Set SuperLib defaults
	this._pf.cost.no_existing_road = repair_existing? 300 : 60; // default = 40
	this._pf.cost.tile = 80; // default = 100
	this._pf.cost.max_bridge_length = 15; // default = 10

	// Set maximum cost as some constant plus a factor times the cost of the optimal path to avoid to long detours
	// factor = 5/2 aka 2,5
	this._pf.cost.max_cost = 7000 + (this._pf.cost.tile * AIMap.DistanceManhattan(sources[0], goals[0]) * 5) / 2 + this._pf.cost.turn;


	this.max_iterations = null;
	this.step_size = null;
	this.accumulated_iterations = 0;
	this.find_path_error_code = _SuperLib_RoadPathFinder.PATH_FIND_NO_ERROR;
}

function _SuperLib_RoadPathFinder::SetMaxIterations(max_iterations)
{
	this.max_iterations = max_iterations;
}

function _SuperLib_RoadPathFinder::SetStepSize(step_size)
{
	this.step_size = step_size;
}

function _SuperLib_RoadPathFinder::FindPath(num_iterations = null)
{
	local num_iterations_to_run = num_iterations;
	if(num_iterations_to_run == null)
	{
		num_iterations_to_run = this.step_size;
	}
		
	// Reduce number of iterations to run if we would run more iterations than maximum allowed
	if(this.accumulated_iterations + num_iterations_to_run > this.max_iterations)
	{
		num_iterations_to_run = this.max_iterations - this.accumulated_iterations;
	}

	if(num_iterations_to_run <= 0)
	{
		// No path found (within time)
		_SuperLib__SuperLib_Log.Info("SuperLib: Finding path took to long -> abort", _SuperLib_Log.LVL_SUB_DECISIONS);
		this.find_path_error_code = _SuperLib_RoadPathFinder.PATH_FIND_FAILED_TIME_OUT;
		return null; 
	}

	// Run FindPath in base class
	local result = this._pf.FindPath(num_iterations_to_run);
	if(result == null)
		this.find_path_error_code = _SuperLib_RoadPathFinder.PATH_FIND_FAILED_NO_PATH;

	return result;
}

function _SuperLib_RoadPathFinder::GetFindPathError()
{
	return this.find_path_error_code;
}

function _SuperLib_RoadPathFinder::GetIterationsUsed()
{
	return this.accumulated_iterations;
}

